from .defaults.tiers import DEFAULT_TIERS
from .environment import Project, env
from .core import TierBase
from cassini.defaults.tiers import Home

try:
    __IPYTHON__  # type: ignore[name-defined]
    from .magics import register

    register()
except NameError:
    pass



import functools

from typing import (
    Callable,
    Union,
    Any,
    cast,
    Dict,
    TypeVar,
    Generic,
    Optional,
    Type,
    overload,
    ClassVar,
)
from typing_extensions import Self

T = TypeVar("T")
V = TypeVar("V")


def _null_func(val):
    return val


JSONType = Union[dict, list, str, int, float, bool, None]
JSONProcessor = Callable[[JSONType], Any]


class MetaAttr:
    """
    Accessor for getting values from a Tier class's meta as an attribute.

    Supports basic serial and de-serialisation.

    Isn't fussy, in that it won't raise an exception if it can't find its key.

    Arguments
    =========
    post_get: func
        function to apply to data after being loaded from json file
    pre_set: func
        function to apply to data before stored in json file.
    name: str
        key to lookup in meta
    default:
        value to return if key not found in meta (note post_get isn't called on this).

    Examples
    ========

    >>> class MyTier(TierBase):
    >>>
    >>>     list_attr = MetaAttr(post_get=lambda val: val.split(','),
    ...                          pre_set=lambda val: ','.join(val))

    """

    def __init__(
        self,
        post_get: JSONProcessor = _null_func,
        pre_set: Callable[[Any], JSONType] = _null_func,
        name: Union[str, None] = None,
        default: Any = None,
    ):
        self.name: str = cast(str, name)
        self.post_get = post_get
        self.pre_set = pre_set
        self.default = default

    def __set_name__(self, owner, name: str):
        if self.name is None:
            self.name = name

    def __get__(self, instance, owner):
        try:
            return self.post_get(instance.meta[self.name])
        except KeyError:
            return self.default

    def __set__(self, instance, value):
        setattr(instance.meta, self.name, self.pre_set(value))


class _SoftProp(Generic[T, V]):
    """
    Like a property, except can be overwritten.
    """

    def __init__(self, func: Callable[[T], V]):
        self.func = func

    @overload
    def __get__(self, instance: None, owner: Type[T]) -> Self:
        pass

    @overload
    def __get__(self, instance: T, owner: Type[T]) -> V:
        pass

    def __get__(self, instance: Optional[T], owner: Type[T]) -> Union[V, Self]:
        if instance:
            return self.func(instance)
        return self


def soft_prop(wraps: Callable[[Any], V]) -> V:
    """
    Create an over-writable property
    """
    return cast(V, functools.wraps(wraps)(_SoftProp(wraps)))


class _CachedProp(Generic[T, V]):
    """
    Like a read only property, except it's only evaluated once.

    Cached value is stored in the _CachedProp instance, which is maybe a bad idea, idk.
    """

    def __init__(self, func: Callable[[T], V]):
        self.func = func
        self.cache: Dict[T, V] = {}

    @overload
    def __get__(self, instance: None, owner: Type[T]) -> Self:
        pass

    @overload
    def __get__(self, instance: T, owner: Type[T]) -> V:
        pass

    def __get__(self, instance: Optional[T], owner: Type[T]) -> Union[V, Self]:
        if instance is None:
            return self

        if instance in self.cache:
            return self.cache[instance]

        val = self.func(instance)
        self.cache[instance] = val

        return val

    def __set__(self, instance, value):
        raise AttributeError("Trying to set a cached property - naughty!")


def cached_prop(wrapped: Callable[[Any], V]) -> V:
    """
    Decorator for turning functions/ methods into `_CachedProp`s.
    """
    return cast(V, functools.wraps(wrapped)(_CachedProp(wrapped)))


class _CachedClassProp(Generic[T, V]):
    """
    Like a combination between `classmethod` and `property`.

    The class instance is passed to the wrapped method upon calling.

    Also performs same caching as `_CachedProp`.
    """

    def __init__(self, func: Callable[[Type[T]], V]):
        self.func = func
        self.cache: Dict[Type[T], V] = {}

    def __get__(self, instance: T, owner: Type[T]) -> V:
        if owner in self.cache:
            return self.cache[owner]

        val = self.func(owner)
        self.cache[owner] = val

        return val

    def __set__(self, instance, value):
        raise AttributeError("Trying to set a cached class property - naughty!")


def cached_class_prop(wrapped: Callable[[Any], V]) -> V:
    """
    Decorator for turning functions/ methods into `_CachedClassProp`s.

    First argument of wrapped will be `self.__class__` rather than `self`.
    """
    return cast(V, functools.wraps(wrapped)(_CachedClassProp(wrapped)))



from typing import List, Union

from IPython.display import publish_display_data

from .environment import Project
from .ipygui import BaseTierGui


class JLGui(BaseTierGui):
    def __init__(self, tier):
        self.tier = tier

    def header(self):
        publish_display_data({"application/cassini.header+json": {}})

    def meta_editor(self, name: Union[str, List[str]]):
        if isinstance(name, str):
            name = [name]

        publish_display_data({"application/cassini.metaeditor+json": {"values": name}})


def extend_project(project: Project):
    for tier_class in project.hierarchy:
        tier_class.gui_cls = JLGui

    return project



import nbformat
import re


class BaseUpdater:
    def cell_processor(self, f):
        self.cell_processors.append(f)
        return f


class V0_1to0_2(BaseUpdater):
    def __init__(self):
        from .. import env

        self.home = env()
        self.cell_processors = []

        @self.cell_processor
        def fix_imports(cell):
            text = cell["source"]
            text = text.replace(
                "from htools.exp import g", "from htools.scify import *"
            )
            cell["source"] = text

        @self.cell_processor
        def fix_g_references(cell):
            text = cell["source"]

            def replace(match):
                return f"env.{match.group(1)}"

            cell["source"] = re.sub("g\.(.+)", replace, text)

        @self.cell_processor
        def replace_mksmpl(cell):
            pattern = "%%mksmpl (.+)"
            text = cell["source"]

            match = re.search(pattern, text)
            if match:
                cell["source"] = f"smpl = env('{match.group(1)}')\nsmpl.header()"

        @self.cell_processor
        def replace_smplconc(cell):
            text = cell["source"]
            cell["source"] = text.replace("%%smplconc", "%%conc")

        @self.cell_processor
        def replace_expconc(cell):
            text = cell["source"]
            cell["source"] = text.replace("%%expconc", "%%conc")

        @self.cell_processor
        def fix_meta_store(cell):
            text = cell["source"]
            pattern = "smpl\.write_meta\('(.+)', (.+)\)"

            def replace(match):
                return f"smpl.{match.group(1)} = {match.group(2)}"

            cell["source"] = re.sub(pattern, replace, text)

    def walk_nbs(self):
        for wp in self.home:
            for exp in wp:
                for smpl in exp:
                    yield smpl

    def update(self):
        for smpl in self.walk_nbs():
            if not smpl.file.exists():
                print(f"No file found for {smpl} - skippiping")
                continue

            print("Fixing", smpl)
            with open(smpl.file, "rb") as f:
                nb = nbformat.read(f, 4)

            print("Backing up old file")
            try:
                new_name = smpl.file.rename(
                    smpl.file.with_name(f"{smpl.name}-old.ipynb")
                )
            except FileExistsError:
                print(f"Already updated {smpl} - skipping")
                continue

            try:
                print("Applying processors")
                for cell in nb["cells"]:
                    for processor in self.cell_processors:
                        processor(cell)
                print("Success")
                print("Errors: ", nbformat.validate(nb, version=4))

                print("Writing new file")
                with open(smpl.file, "w", encoding="utf-8") as f:
                    nbformat.write(nb, f, 4)
                print("Success")
            except Exception:
                print("Exception occured, rolling back")
                new_name.rename(smpl.file.name)



from pathlib import Path
import os
import sys
import functools
from typing import Union, Callable, Any
from typing_extensions import Self
import datetime

from .config import config
from .accessors import JSONType


def str_to_date(val: JSONType) -> datetime.datetime:
    assert isinstance(val, str)
    return datetime.datetime.strptime(val, config.DATE_FORMAT)


def date_to_str(date: datetime.datetime) -> str:
    return date.strftime(config.DATE_FORMAT)


class FileMaker:
    """
    Utility class for making files, and then rolling back if an exception occurs.
    """

    def __init__(self):
        self.folders_made = []
        self.files_made = []

    def __enter__(self):
        self.files_made = []
        self.folders_made = []
        return self

    def mkdir(self, path: Path, exist_ok: bool = False) -> Union[Path, None]:
        if not path.exists():
            path.mkdir()
            self.files_made.append(path)
            return path
        if exist_ok:
            return None
        raise FileExistsError(path)

    def write_file(
        self, path: Path, contents: str = "", exist_ok: bool = False
    ) -> Union[Path, None]:
        if not path.exists():
            path.write_text(contents)
            self.files_made.append(path)
            return path
        if exist_ok:
            return None
        raise FileExistsError(path)

    def copy_file(self, source, dest):
        self.write_file(dest, source.read_text())
        return source, dest

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            print(f"{exc_type} occured, rolling back")
            for file in self.files_made:
                print("Deleting", file)
                file.unlink()
                print("Done")

            for folder in self.folders_made:
                print("Removing", folder)
                folder.rmdir()
                print("Done")
            raise exc_type(exc_val)


class XPlatform:
    """
    Class for easily making cross platform functions/ methods(?). Stops nasty if elses.
    """

    def __init__(self) -> None:
        self._func: Union[Callable, None] = None
        self._default: Union[Callable, None] = None

    @property
    def func(self) -> Callable:
        """
        Returns the platform appropriate function (or default if not known)
        """
        if self._func:
            return self._func

        assert self._default

        return self._default

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def add(self, platform: str) -> Callable[[Any], Self]:
        """
        Add a platform dependent function. To be used as a decorator.

        Parameters
        ----------
        platform: str
            platform wrapped function called for.

        Returns
        -------
        self : XPlatform
            Callable that will always call the appropriate function for the current platform.

        """

        def wrapper(func):
            if sys.platform == platform:
                self._func = func
                functools.update_wrapper(self, func)
            return self

        return wrapper

    def default(self, func: Callable) -> "XPlatform":
        """
        Fallback functional called if no matching function added for current platform.

        Parameters
        ----------
        func: Callable

        Returns
        -------
        self: XPlatform
            Callable that will always call the appropriate function for the current platform.
        """
        self._default = func
        return self


open_file = XPlatform()


@open_file.add("win32")
def win_open_file(filename: str):
    """
    Windows open file implementation.
    """
    # mypy doesn't understand XPlatform implementation... maybe this is working too hard to please?
    if sys.platform != "win32":
        raise RuntimeError("Calling win32 specific method in wrong platform")

    os.startfile(filename)


@open_file.default
def nix_open_file(filename: str):
    """
    *Nix open file implementation.
    """
    import subprocess

    opener = "open" if sys.platform == "darwin" else "xdg-open"
    subprocess.call([opener, filename])



import html

from typing import Any, Callable, Dict, Union, List, Mapping, Hashable, TYPE_CHECKING

import pandas as pd
import pandas.api.types as pd_types

from ipywidgets import Select, VBox, Button, Output, Text, Textarea, HBox, Layout, Accordion, DOMWidget  # type: ignore[import]
from IPython.display import display, Markdown, publish_display_data

from .environment import env

if TYPE_CHECKING:
    from .core import TierBase


class WHTML:
    """
    Magic class that takes html text and creates an object that is rendered with that text.

    `ipywidget.HTML` widgets do not get their links XSRF protected, but `obj._repr_html_` does!

    Warnings
    --------
    This HTML is not escaped, so could do bad stuff!
    """

    def __init__(self, html: str):
        self.html = html

    def _repr_html_(self):
        return self.html


def widgetify(obj: Any) -> Output:
    """
    Allows any object to be treated like an `ipywidget` so it can be used in `ipywidget` layouts - handy!
    """
    output = Output(height="auto", width="auto")
    with output:
        display(obj)
    return output


def widgetify_html(html: str) -> Output:
    """
    Required for html which has links rather than ipywidgets HTML widget.

    Parameters
    ----------
    html: str
        HTML to render

    Returns
    -------
    Output widget containing the HTML
    """
    return widgetify(WHTML(html))


class UnescapedDataFrame(pd.DataFrame):
    """
    Subclass of `pd.DataFrame` that slightly dangerously does not escape any html in its `self._repr_html_` method.

    Warnings
    --------
    This class deliberately does not escape HTML for objects it contains. This means you need to trust these objects.
    """

    @staticmethod
    def try_html_repr(obj: object) -> str:
        """
        Get text form of obj, first try `obj._repr_html_`, fallback on `str(obj)`.

        Warnings
        --------
        This method does not escape `obj._repr_html_`. So make sure your reprs are trustworthy!
        """
        if hasattr(obj, "_repr_html_"):
            return obj._repr_html_()
        return html.escape(str(obj))

    @property
    def _constructor(self):
        return UnescapedDataFrame

    def _repr_html_(self) -> str:
        formatters: Mapping[Hashable, Callable[[object], str]] = {
            name: self.try_html_repr
            for name, dtype in self.dtypes.items()
            if pd_types.is_object_dtype(dtype)
        }
        return self.to_html(escape=False, formatters=formatters)


class InputSequence:
    """
    Helper class for creating 'input sequences' i.e. a sequency of widgets that you fill in, and then click confirm.

    Parameters
    ----------
    done_action : callable
        function to call when confirm is clicked. Each child (see next parameter!) is passed to done_action
    *children : ipywidget
        widgets to display in order. When confirm clicked, these widget objects are passed to done_action.
    """

    stored_traits = ["value", "disabled"]

    def __init__(self, done_action: Callable, *children: DOMWidget):
        self.done_action = done_action
        self.children = children
        self.initial_traits = {}

        self.confirm = Button(description="Confirm", disabled=True)
        self.new = Button(description="New")
        self.status = Output()

        self.output = Output(layout=Layout(width="auto"))

        self.confirm.on_click(self._do_done)
        self.new.on_click(self.reset)

        for child in self.children:
            child.layout.width = "auto"
            child.style.description_width = "10%"
            self.initial_traits[child] = {
                k: v for k, v in child.get_state().items() if k in self.stored_traits
            }
            child.observe(self.child_updated, "value")

    def reset(self, change):
        """
        Reset the state of the InputSequence. Called when `new` button clicked.
        """
        for child in self.children:
            child.open()
            traits = self.initial_traits[child]
            for k, v in traits.items():
                setattr(child, k, v)

        self.confirm.disabled = True

    def child_updated(self, change):
        """
        Called when a child is updated, meant to stop you pressing confirm until you've entered something into every
        widget... don't think it actually works tho!
        """
        if all(map(lambda val: val is not None, self.children)):
            self.confirm.disabled = False

    def as_widget(self) -> DOMWidget:
        """
        Build `self` into `ipywidget`

        Returns
        -------
        widget : DOMWidget
            widget form of self.
        """
        return VBox((*self.children, HBox((self.confirm, self.new)), self.status))

    def display(self):
        """
        Displays `self.as_widget()`.

        Returns
        -------
        output : Output
            output widget which we displayed into...
        """
        self.output.append_display_data(self.as_widget())
        return self.output

    def _do_done(self, pressed):
        """
        Internal function called when confirm clicked.
        """
        values = []

        self.confirm.disabled = True

        for child in self.children:
            child.disabled = True
            values.append(child.value)

        with self.output:
            self.done_action(*values)


class SearchWidget:
    def __init__(self):
        self.search = Text(placeholder="Name")
        self.go_btn = Button(description="Search")
        self.clear_btn = Button(description="Clear")
        self.out = Output()

        self.search.continuous_update = False
        self.search.observe(self.do_search, "value")
        self.go_btn.on_click(self.do_search)
        self.clear_btn.on_click(self.do_clear)

    def do_clear(self, cb):
        self.out.clear_output()

    def do_search(self, cb):
        name = self.search.value

        self.out.clear_output()

        with self.out:
            obj = env.project[name]
            if obj.exists():
                display(obj.gui.header())
            else:
                print(f"{obj.name} valid, but doesn't exist")

    def as_widget(self):
        return VBox([HBox([self.search, self.go_btn, self.clear_btn]), self.out])


class BaseTierGui:
    """
    Mixin to provide nice notebook outputs for Jupyter Notebooks.
    """

    def __init__(self, tier: "TierBase"):
        self.tier = tier

    def _get_header_components(self) -> Dict[str, Callable[[], DOMWidget]]:
        """
        Creates the dictionary that forms the components of the header.

        Returns
        -------
        components : dict
            Dictionary of callables in form (key: value) = (header, callable), where the callable creates a returns a
            widget.

        Notes
        -----
        The order of the elements dictates what order they're rendered so insert them in the order you want!

        Overload this method to customise the appearance of your `Tier` header.
        """
        components: Dict[str, Callable[[], Any]] = dict()
        components["Description"] = self._build_description
        components["Highlights"] = self._build_highlights_accordion
        components["Conclusion"] = self._build_conclusion

        child = self.tier.child_cls
        if child:
            child_name = child.pretty_type
            components[f"{child_name}s"] = self._build_children
            components[f"New {child_name}"] = self.new_child

        return components

    def _build_header_title(self, *, parent=None) -> DOMWidget:
        """
        Creates a widget that displays the title a `Tier` along side clickable links and buttons to its components
        """
        open_btn = Button(description="Folder")
        open_btn.on_click(lambda cb: self.tier.open_folder())
        title = f'<h1 style="display: inline;"><a href="{self.tier.href}">{self.tier.name}</a></h1>'
        parts = []
        parent = self.tier.parent
        while parent:
            parts.append(f'<a href="{parent.href}" target="_blank">{parent.name}</a>')
            parent = parent.parent

        text = title + '<h3 style="display: inline;">'

        if parts:
            text += f"({'->'.join(parts[::-1])})"

        text += "</h3>"
        return VBox((widgetify_html(text), open_btn))

    def _build_description(self, *, parent=None) -> Union[DOMWidget, None]:
        """
        Creates a widget that displays the motivation for a `Tier`.
        """
        description = self.tier.description
        if not description:
            return None
        return widgetify(Markdown(description))

    def _build_highlights_accordion(self, *, parent=None) -> Union[DOMWidget, None]:
        """
        Creates a widget that displays highlights for this `Tier` in an `ipywidgets.Accordion` - which is nice!
        """
        highlights = self.tier.get_highlights()

        if not highlights:
            return None

        widget = Accordion()

        for i, (name, highlight) in enumerate(highlights.items()):
            out = Output()
            widget.children = (*widget.children, out)
            with out:
                for item in highlight:
                    publish_display_data(**item)
            widget.set_title(i, name)
        widget.selected_index = None
        return widget

    def _build_conclusion(self, *, parent=None) -> Union[DOMWidget, None]:
        """
        Build widget to display conclusion of this `Tier` object.
        """
        conclusion = self.tier.conclusion
        if not conclusion:
            return None
        return widgetify(Markdown(self.tier.conclusion))

    def _build_children(self, *, parent=None) -> DOMWidget:
        """
        Build a widget to display an `UnescapedDataFrame` containing this `Tier`'s children.
        """
        children_df = self.children_df()
        return widgetify(children_df)

    def header(
        self,
        *,
        include: Union[List[str], None] = None,
        exclude: Union[List[str], None] = None,
    ) -> DOMWidget:
        """
        Builds header widget from its components.

        Parameters
        ----------
        include : Sequence[str]
            names of components to render in header widget.
        exclude : Sequence[str]
            names of components not to render in header widget.

        Notes
        -----
        Parameters include and exclude are mutually exclusive.

        Returns
        -------
        header : DOMWidget
            Widget that can be displayed containing all the components of this `Tier`'s header.
        """
        if include and exclude:
            raise ValueError("Only one of include or exclude can be provided")

        component_makers = self._get_header_components()

        if include:
            component_makers = {
                k: v for k, v in component_makers.items() if k in include
            }

        if exclude:
            component_makers = {
                k: v for k, v in component_makers.items() if k not in exclude
            }

        components = [self._build_header_title()]

        for name, component_builder in component_makers.items():
            component = component_builder()
            if component:
                components.append(widgetify_html(f"<h3>{name}</h3>"))
                components.append(component)

        return VBox(components)

    def display_highlights(self):
        """
        Display an `ipywidgets.Accordian` with this `Tier`'s highlights.
        """
        return display(self._build_highlights_accordion())

    def children_df(
        self,
        *,
        include: Union[List[str], None] = None,
        exclude: Union[List[str], None] = None,
    ) -> Union[UnescapedDataFrame, None]:
        """
        Calls `tier.children_df` but returns an `UnescapedDataFrame` instead.
        """
        return UnescapedDataFrame(
            self.tier.children_df(include=include, exclude=exclude)
        )

    def new_child(self, *, parent=None) -> DOMWidget:
        """
        Widget for creating new child for this `Tier`.
        """
        child = self.tier.child_cls

        if not child:
            return

        options = child.get_templates()

        mapping = {path.name: path for path in options}

        selection = Select(options=mapping.keys(), description="Template")

        def create(name, template, description):
            with form.status:
                obj = child(*self.tier.identifiers, name)
                obj.setup_files(mapping[template])
                obj.description = description
                display(widgetify_html(obj._repr_html_()))
            parent.update("Children")

        form = InputSequence(
            create,
            Text(description=f"Identifier", placeholder=f"{child.id_regex}"),
            selection,
            Textarea(description="Motivation"),
        )

        return form.as_widget()



from __future__ import annotations

from pathlib import Path
from warnings import warn
import re
import os

from jupyterlab.labapp import LabApp  # type: ignore
from jupyterlab_server import LabConfig

from typing import Union, Tuple, List, TYPE_CHECKING, Type, TypeVar
from typing_extensions import TypeGuard

if TYPE_CHECKING:
    from .core import TierBase

import jinja2

from .accessors import soft_prop
from .utils import FileMaker
from .config import config


class PathLibEnv(jinja2.Environment):
    """
    Subclass of `jinja2.Environment` to enable using `pathlib.Path` for template names.
    """

    def get_template(self, name: Union[Path, str], parent=None, globals=None):  # type: ignore[override]
        return super().get_template(
            name.as_posix() if isinstance(name, Path) else name,
            parent=None,
            globals=None,
        )


class CassiniLabApp(LabApp):
    """
    Subclass of `jupyterlab.labapp.LabApp` that ensures `ContentsManager.allow_hidden = True` (needed for jupyter_cassini_server)
    """

    @classmethod
    def initialize_server(cls, argv=None):
        """
        Patch serverapp to ensure hidden files are allowed, needed for jupyter_cassini_server
        """
        serverapp = super().initialize_server(argv)
        serverapp.contents_manager.allow_hidden = True
        return serverapp


class Project:
    """
    Represents your project. Understands your naming convention, and your project hierarchy.

    Parameters
    ----------
    hierarchy : List[Type[BaseTier]]
        Sequence of `TierBase` subclasses representing the hierarchy for this project. i.e. earlier entries are stored
        in higher level directories.
    project_folder : Union[str, Path]
        path to home directory. Note this also accepts a path to a file, but will take `project_folder.parent` in that
        case. This enables `__file__` to be used if you want `project_folder` to be based in the same dir.

    Notes
    -----
    This class is a singleton i.e. only 1 instance per interpreter can be created.
    """

    _instance: Union[Project, None] = None

    def __new__(cls, *args, **kwargs):
        if cls._instance:
            raise RuntimeError(
                "Attempted to create new Project instance, only 1 instance permitted per interpreter"
            )
        instance = object.__new__(cls)
        cls._instance = instance
        env.project = instance
        return instance

    def __init__(
        self, hierarchy: List[Type[TierBase]], project_folder: Union[str, Path]
    ):
        self.rank_map = {}
        self.hierarchy = hierarchy

        project_folder = Path(project_folder).resolve()
        self.project_folder = (
            project_folder if project_folder.is_dir() else project_folder.parent
        )

        self.template_env = PathLibEnv(
            autoescape=jinja2.select_autoescape(["html", "xml"]),
            loader=jinja2.FileSystemLoader(self.template_folder),
        )

        for rank, tier_cls in enumerate(hierarchy):
            tier_cls.rank = rank
            self.rank_map[tier_cls.short_type] = rank

    @property
    def home(self):
        """
        Get the home `Tier`.
        """
        return self.hierarchy[0]()

    def env(self, name: str) -> TierBase:
        """
        Initialise the global environment to a particular `Tier` that is retrieved by parsing `name`.

        This will set the value of `env.o`.

        Warnings
        --------
        This should only really be called once (or only with 1 name). Otherwise this could create some unexpected
        behaviour.
        """
        obj = self.__getitem__(name)

        if env.o and name != env.o.name:
            warn(
                (
                    f"Overwriting the global Tier {env.o} for this interpreter. This may cause unexpected behaviour. "
                    f"If you wish to create Tier objects that aren't the current Tier I recommend initialising them "
                    f"directly e.g. obj = MyTier('id1', 'id2')"
                )
            )

        env.update(obj)
        return obj

    def __getitem__(self, name: str) -> TierBase:
        """
        Retrieve a tier object from the project by name.

        Parameters
        ----------
        name : str
            Parsable name to get the tier object by. To get your `Home` just provide `Home.name`.

        Returns
        -------
        tier : TierBase
            Tier retrieved from project.
        """
        if name == self.home.name:
            obj = self.home
        else:
            identifiers = self.parse_name(name)
            if not identifiers:
                raise ValueError(f"Name {name} not recognised as identifying any Tier")
            cls = self.hierarchy[len(identifiers)]
            obj = cls(*identifiers)
        return obj

    @soft_prop
    def template_folder(self):
        """
        Overwritable property providing where templates will be stored for this project.
        """
        return self.project_folder / "templates"

    @soft_prop
    def caslib_folder(self) -> Path:
        return self.project_folder / "caslib"

    def setup_files(self):
        """
        Setup files needed for this project.

        Will put everything you need in `project_folder` to get going.
        """
        home = self.home
        if home.exists():
            return home

        print("Setting up project.")

        with FileMaker() as maker:
            print(f"Creating templates folder")
            maker.mkdir(self.template_folder)
            print("Success")

            for tier_cls in self.hierarchy:
                if tier_cls.default_template is None:
                    continue
                maker.mkdir(self.template_folder / tier_cls.pretty_type)
                print("Copying over default template")
                maker.copy_file(
                    config.BASE_TEMPLATE,
                    self.template_folder / tier_cls.default_template,
                )
                print("Done")

        print(f"Setting up Home Tier")
        home.setup_files()
        print("Success")

    def launch(self, app=None, patch_pythonpath=True):
        """
        Jump off point for a cassini project.

        Sets up required files for your project, monkeypatches `PYTHONPATH` to make your project available throughout
        and launches a jupyterlab server.

        This explicitly associates an instance of the Jupyter server with a particular project.

        Parameters
        ----------
        app : LabApp
            A ready made Jupyter Lab app (By defuault will just create a new one).
        patch_pythonpath : bool
            Add `self.project_folder` to the `PYTHONPATH`? (defaults to `True`)
        """
        self.setup_files()

        if patch_pythonpath:
            py_path = os.environ.get("PYTHONPATH", "")
            project_path = str(self.project_folder.resolve())
            os.environ["PYTHONPATH"] = (
                py_path + os.pathsep + project_path if py_path else project_path
            )

        if app is None:
            app = CassiniLabApp()

        app.launch_instance()

        return app

    def parse_name(self, name: str) -> Union[Tuple[str, ...], Tuple]:
        """
        Parses a string that corresponds to a `Tier` and returns a list of its identifiers.

        returns an empty tuple if not a valid name.

        Parameters
        ----------
        name : str
            name to parse

        Returns
        -------
        identifiers : tuple
            identifiers extracted from name, empty tuple if `None` found

        Notes
        -----
        This works in a slightly strange - but robust way!

        e.g.

            >>> name = 'WP2.3c'

        it will loop through each entry in `cls.hierarchy` (skipping home!), and then perform a search on `name` with
        that regex:

            >>> WorkPackage.name_part_regex
            WP(\\d+)
            >>> match = re.search(WorkPackage.name_part_regex, name)

        If there's no match, it will return `()`, if there is, it stores the `id` part:

            >>> wp_id = match.group(1)  # in python group 0 is the whole match
            >>> wp_id
            2

        Then it removes the whole match from the name:

            >>> name = name[match.end(0):]
            >>> name
            .3c

        Then it moves on to the next tier

            >>> Experiment.name_part_regex
            '\\.(\\d+)'
            >>> match = re.search(WorkPackage.name_part_regex, name)

        If there's a match it extracts the id, and substracts the whole string from name and moves on, continuing this
        loop until it's gone through the whole hierarchy.

        The whole name has to be a valid id, or it will return `()` e.g.

            >>> TierBase.parse_name('WP2.3')
            ('2', '3')
            >>> TierBase.parse_name('WP2.u3')
            ()
        """
        parts = self.hierarchy[1:]
        ids: List[str] = []
        for tier_cls in parts:
            pattern = tier_cls.name_part_regex
            match = re.search(pattern, name)
            if match and match.start(0) == 0:
                ids.append(match.group(1))
                name = name[match.end(0) :]
            else:
                break
        if name:  # if there's any residual text then it's not a valid name!
            return tuple()
        else:
            return tuple(ids)

    def __repr__(self):
        return f"<Project at: '{self.project_folder}' hierarchy: '{self.hierarchy}' ({env})>"


ValWithInstance = TypeVar("ValWithInstance")


class _Env:
    """
    Essentially a global object that describes the state of the project for this interpreter.

    As each notebook has its own interpreter, each notebook also has its own env that basically stores what the current
    project is and what tier this `ipynb` file corresponds to.

    Attributes
    ----------
    project : Project
        reference to the current project object. Returns `None` if one not set yet.

    Warnings
    --------
    This object is a singleton, so only one instance can exist at a time.

    This object shouldn't be created directly, instead you should call `project.env('...')` to set its value.
    """

    instance: Union[_Env, None] = None

    def __new__(cls, *args, **kwargs):
        if cls.instance:
            raise RuntimeError(
                "Attempted to create new _Env instance, only 1 _instance permitted per interpreter"
            )
        instance = object.__new__(cls)
        cls.instance = instance
        return instance

    def __init__(self) -> None:
        self.project: Union[Project, None] = None
        self._o: Union[TierBase, None] = None

    def _has_instance(
        self, val: Union[ValWithInstance, None]
    ) -> TypeGuard[ValWithInstance]:
        return self.instance is not None

    @property
    def o(self) -> Union[TierBase, None]:
        """
        Reference to current Tier object.
        """
        if self._has_instance(self._o):
            return self._o

        return None

    def update(self, obj: TierBase):
        self._o = obj


env = _Env()



from pathlib import Path
import os
import html

from typing import Iterator, List, overload

import pandas as pd  # type: ignore[import]
from IPython.display import display
from ipywidgets import SelectMultiple, Text, HBox, Button  # type: ignore[import]

from ..core import TierBase
from ..accessors import cached_prop, cached_class_prop
from ..utils import FileMaker
from ..ipygui import InputSequence, widgetify_html, BaseTierGui, SearchWidget
from ..environment import env


def ignore_dir(name):
    if name.startswith("."):
        return True
    if name.startswith("_"):
        return True
    return False


class HomeGui(BaseTierGui):
    def _get_header_components(self):
        components = dict()
        components["Search"] = lambda: SearchWidget().as_widget()
        child = self.tier.child_cls
        if child:
            child_name = child.pretty_type
            components[f"{child_name}s"] = self._build_children
            components[f"New {child_name}"] = self.new_child

        return components


class Home(TierBase):
    """
    Home `Tier`.

    This, or a subclass of this should generally be the first entry in your hierarchy, essentially represents the top
    level folder in your hierarchy.

    Creates the `Home.ipynb` notebook that allows easy navigation of your project.
    """

    gui_cls = HomeGui

    @cached_prop
    def name(self):
        return self.pretty_type

    @cached_prop
    def folder(self):
        assert env.project
        assert self.child_cls
        return env.project.project_folder / (self.child_cls.pretty_type + "s")

    @cached_prop
    def file(self):
        assert env.project
        return env.project.project_folder / f"{self.name}.ipynb"

    @cached_prop
    def highlights_file(self):
        return None

    @cached_prop
    def meta_file(self):
        return None

    def serialize(self) -> dict:
        data = {}

        data["identifiers"] = self.name
        data["name"] = self.name
        data["file"] = str(self.file)
        data["parents"] = []
        data["children"] = [child.name for child in self]

        return data

    def exists(self):
        return self.file.exists()

    def setup_files(self):
        assert self.child_cls

        with FileMaker() as maker:
            print(f"Creating {self.child_cls.pretty_type} folder")
            maker.mkdir(self.folder)
            print("Success")

        with FileMaker() as maker:
            print(f"Creating Tier File ({self.file})")
            maker.write_file(self.file, self.render_template(self.default_template))
            print("Success")


class WorkPackage(TierBase):
    """
    WorkPackage Tier.

    Intended to contain all the work towards a particular goal. i.e. prove that we can do this.

    Next level down are `Experiment`s.
    """

    name_part_template = "WP{}"

    short_type = "wp"

    @property
    def exps(self):
        """
        Gets a list of all this `WorkPackage`s experiments.
        """
        return list(self)


class ExperimentGui(BaseTierGui):
    def new_dataset(self):
        """
        A handy widget for creating new `DataSets`.
        """
        samples = list(self.tier)
        option_map = {sample.name: sample for sample in samples}

        selection = SelectMultiple(options=option_map.keys(), description="Auto Add")

        def create(name, auto_add):
            with form.status:
                self.tier.setup_technique(name)
                if auto_add:
                    for sample in (option_map[name] for name in auto_add):
                        o = sample[name]
                        o.setup_files()
                        display(widgetify_html(o._repr_html_()))

        form = InputSequence(
            create, Text(description=f"Name:", placeholder="e.g. XRD"), selection
        )

        return form.as_widget()

    def _get_header_components(self):
        components = super()._get_header_components()
        components["New Data"] = self.new_dataset
        return components


class Experiment(TierBase):
    """
    Experiment `Tier`.

    Just below `WorkPackage`, experiments are intended to be collections of samples and datasets that work towards the
    goal of the parent `WorkPackage`.

    Each `Experiment` has a number of samples.
    """

    name_part_template = ".{}"

    short_type = "exp"

    gui_cls = ExperimentGui

    @property
    def techniques(self) -> List[str]:
        """
        Convenience property for looking up all the techniques that have been performed on samples in this experiment.

        Notes
        -----
        This just checks for the existence of DataSet folders, and not if they have anything in them!
        """
        techs = []
        for entry in os.scandir(self.folder):
            if entry.is_dir() and not ignore_dir(entry.name):
                techs.append(entry.name)
        return techs

    def setup_technique(self, name):
        """
        Convenience method for adding a new technique to this experiment.

        Essentially just creates a new folder for it in the appropriate location.

        This folder can then be filled with `DataSet`s
        """
        print("Making Data Folder")
        folder = self.folder / name

        if folder.exists():
            raise FileExistsError(f"{folder} exists already")

        with FileMaker() as maker:
            maker.mkdir(folder)

        print("Done")

    def children_df(self, include=None, exclude=None):
        df = super().children_df(include=include, exclude=exclude)

        if df is None:
            return None

        df["datasets"] = pd.Series(
            {smpl.name: list(dataset.id for dataset in smpl.datasets) for smpl in self}
        )
        return df

    @property
    def smpls(self):
        """
        Get a list of this `Experiment`s samples.
        """
        return list(self)


class SampleGui(BaseTierGui):
    def new_child(self):
        def create(name):
            with form.status:
                o = self.tier[name]
                o.setup_files()

        form = InputSequence(create, Text(description=f"Name:", placeholder="e.g. XRD"))

        return form.as_widget()

    def _build_children(self):
        buttons = []
        for dataset in self.tier.datasets:
            b = Button(description=dataset.id)

            def make_callback(dataset):
                def open_folder(change):
                    dataset.open_folder()

                return open_folder

            b.on_click(make_callback(dataset))
            buttons.append(b)
        return HBox(tuple(buttons))

    def _get_header_components(self):
        components = super()._get_header_components()
        return components


class Sample(TierBase):
    """
    Sample `Tier`.

    A `Sample` is intended to represent some object that you collect data on.

    As such, each sample has its own `DataSet`s.

    Notes
    -----
    A `Sample` id can't start with a number and can't contain `'-'` (dashes), as these confuse the name parser.
    """

    name_part_template = "{}"
    id_regex = r"([^0-9^-][^-]*)"

    gui_cls = SampleGui

    @cached_prop
    def folder(self):
        assert self.parent
        return self.parent.folder

    @property
    def datasets(self) -> list:
        """
        Convenient way of getting a list of `DataSet`s this sample has.
        """
        assert self.parent
        assert self.child_cls

        techs = []
        for technique in self.parent.techniques:
            dataset = self.child_cls(*self.identifiers, technique)
            if dataset.exists():
                techs.append(dataset)
        return techs

    def __iter__(self):
        return iter(self.datasets)


class DataSet(TierBase):
    """
    `DataSet` Tier.

    The final tier, intended to represent a folder containing a collection of files relating to a particular `Sample`.
    """

    short_type = "dset"

    name_part_template = "-{}"
    id_regex = r"(.+)"

    @cached_class_prop
    def default_template(cls):
        return None

    @cached_prop
    def folder(self):
        assert self.parent

        return self.parent / self.id / self.parent.id

    @cached_prop
    def href(self):
        return (
            html.escape(Path(os.path.relpath(self.folder, os.getcwd())).as_posix())
            + "/"
        )

    def exists(self):
        return self.folder.exists()

    def setup_files(self, template=None):
        print(f"Creating Folder for Data: {self}")

        with FileMaker() as maker:
            maker.mkdir(self.folder.parent, exist_ok=True)
            maker.mkdir(self.folder)

        print(f"Success")

    @cached_prop
    def meta_file(self):
        """
        `DataSet`s have no meta.
        """
        return None

    @cached_prop
    def highlights_file(self):
        """
        `DataSet`s have no highlights.
        """
        return None

    @cached_prop
    def file(self):
        """
        `DataSet`s have no file
        """
        return None

    @classmethod
    def get_templates(cls):
        """
        Datasets have no templates.
        """
        return []

    def __truediv__(self, other):
        return self.folder / other

    def __iter__(self) -> Iterator[os.DirEntry]:
        """
        Call `os.scandir` on `self.folder`.
        """
        yield from os.scandir(self.folder)

    def __fspath__(self):
        return self.folder.__fspath__()


DEFAULT_TIERS = [Home, WorkPackage, Experiment, Sample, DataSet]



from IPython.core.magic import register_cell_magic
from IPython.core.interactiveshell import InteractiveShell
from IPython.display import Markdown, publish_display_data

from .environment import env


def hlt(line, cell):
    if not line:
        raise ValueError("Please provide a title for the highlight")
    outputs = []

    def capture_display(msg):
        if msg["content"]["data"] in outputs:  # display only once
            return None
        outputs.append(msg["content"])
        return msg

    cell = cell.strip()

    annotation = None

    if cell.endswith('"""'):
        *rest, annotation = cell.split('"""')[:-1]
        cell = '"""'.join(rest)

    shell = InteractiveShell.instance()
    # Capture any output that is displayed before output (like matplotlib plots)
    shell.display_pub.register_hook(capture_display)

    header = f"## {line}"

    try:
        publish_display_data({"text/markdown": header})

        result = shell.run_cell(cell).result
        outputs.append(
            dict(zip(("data", "metadata"), shell.display_formatter.format(result)))
        )

        if annotation:
            publish_display_data({"text/markdown": annotation})
    finally:
        shell.display_pub.unregister_hook(capture_display)

    all_out = outputs

    env.o.add_highlight(line, all_out)

    return None


def cache(line, cell):
    key = str(hash(cell.strip()))
    cache = env.o.get_cached().get(key)

    if cache:
        print(f"Using cached version ('{key}')")
        for output in cache:
            publish_display_data(**output)
        return None

    outputs = []

    def capture_display(msg):
        if msg["content"]["data"] in outputs:  # display only once
            return None
        outputs.append(msg["content"])
        return msg

    shell = InteractiveShell.instance()
    # Capture any output that is displayed before output (like matplotlib plots)
    shell.display_pub.register_hook(capture_display)

    try:
        output = shell.run_cell(cell)
        result = output.result
        if output.error_in_exec:
            raise RuntimeError("Error in cell output, not caching")
        outputs.append(
            dict(zip(("data", "metadata"), shell.display_formatter.format(result)))
        )
    finally:
        shell.display_pub.unregister_hook(capture_display)

    all_out = outputs

    env.o.cache_result(key, all_out)

    return None


def conc(line, cell):
    if env.o.conclusion and line != "force":
        raise RuntimeError(
            f"Conclusion for {env.o} already set, use %%conc force to force update"
        )
    env.o.conclusion = cell
    return Markdown(cell)


def register():
    register_cell_magic(hlt)
    register_cell_magic(cache)
    register_cell_magic(conc)



from pathlib import Path
from dataclasses import dataclass


SCIFY_DIR = Path(__file__).parent


@dataclass
class Config:
    SCIFY_DIR = SCIFY_DIR
    DATE_FORMAT = "%d/%m/%Y"
    META_DIR_TEMPLATE = ".{}s"

    DEFAULT_TEMPLATE_DIR = SCIFY_DIR / "defaults" / "templates"
    TEMPLATE_EXT = ".tmplt.ipynb"
    BASE_TEMPLATE = DEFAULT_TEMPLATE_DIR / "Tier.tmplt.ipynb"


config = Config()



from __future__ import annotations

import datetime
import html
import json
import os
import time
from pathlib import Path
from collections import defaultdict

from typing import (
    Any,
    KeysView,
    List,
    Type,
    Tuple,
    Iterator,
    Union,
    Dict,
    ClassVar,
    Sequence,
    Optional,
    cast,
)

import pandas as pd  # type: ignore

from .ipygui import BaseTierGui
from .accessors import MetaAttr, cached_prop, cached_class_prop, _CachedClassProp
from .utils import FileMaker, open_file, str_to_date, date_to_str
from .environment import env, _Env, Project
from .config import config


class Meta:
    """
    Like a dictionary, except linked to a json file on disk. Caches the value of the json in itself.

    Arguments
    ---------
    file: Path
           File Meta object stores information about.
    """

    timeout: ClassVar[int] = 1
    my_attrs: ClassVar[List] = ["_cache", "_cache_born", "file"]

    def __init__(self, file: Path):
        self._cache: dict = {}
        self._cache_born: float = 0.0
        self.file: Path = file

    @property
    def age(self) -> float:
        """
        time in secs since last fetch
        """
        return time.time() - self._cache_born

    def fetch(self) -> dict:
        """
        Fetches values from the meta file and updates them into `self._cache`.

        Notes
        -----
        This doesn't *overwrite* `self._cache` with meta contents, but updates it. Meaning new stuff to file won't be
        overwritten, it'll just be loaded.
        """
        if self.file.exists():
            self._cache.update(json.loads(self.file.read_text()))
            self._cache_born = time.time()
        return self._cache

    def refresh(self):
        """
        Check age of cache, if stale then re-fetch
        """
        if self.age >= self.timeout:
            self.fetch()

    def write(self):
        """
        Overwrite contents of cache into file
        """
        # Danger moment - writing bad cache to file.
        with self.file.open("w") as f:
            json.dump(self._cache, f)

    def __getitem__(self, item: str) -> Any:
        self.refresh()
        return self._cache[item]

    def __setitem__(self, key: str, value: Any):
        self.__setattr__(key, value)

    def __getattr__(self, item: str) -> Any:
        self.refresh()
        try:
            return self[item]
        except KeyError:
            raise AttributeError(item)

    def __setattr__(self, name: str, value: Any) -> Any:
        if name in self.my_attrs:
            super().__setattr__(name, value)
        else:
            self.fetch()
            self._cache[name] = value
            self.write()

    def __delitem__(self, key: str):
        self.fetch()
        del self._cache[key]
        self.write()

    def __repr__(self) -> str:
        self.refresh()
        return f"<Meta {self._cache} ({self.age * 1000:.1f}ms)>"

    def get(self, key: str, default: Any = None) -> Any:
        """
        Like `dict.get`
        """
        try:
            return self.__getattr__(key)
        except AttributeError:
            return default

    def keys(self) -> KeysView:
        """
        like `dict.keys`
        """
        self.refresh()
        return self._cache.keys()


class TierMeta(type):
    """
    Metaclass mixin for the TierBase. Needed to ensure that each Tier Class has its own cache.
    """

    cache: Dict[Tuple[str, ...], TierBase]

    def __new__(cls, name, bases, dct):
        kls = super().__new__(cls, name, bases, dct)
        kls.cache = {}  # ensures each TierBase class has its own cache
        return kls


class TierBase(metaclass=TierMeta):
    """
    Base class for creating Tiers

    Parameters
    ----------
    *identifiers : str
        sequence of strings that identify this tier. With the final identifier being unique.

    Attributes
    ----------
    rank : int
        Class attribute, specifying the rank of this `Tier`
    id_regex: str
        Class attribute, regex that defines a group that matches the id of a `Tier` object from a name... except the name isn't the full
        name, but with parent names stripped, see examples basically!
    hierarchy : list
        Class attribute, hierarchy of `Tier`s.
    description : str
        returns the description found in a `Tier` _instance's meta file
    started : datetime
        return a datetime parsed using `config.DATE_FORMAT` found in meta file
    conclusion : str
        return the conclusion found in a `Tier` _instance's meta file.
    rank : int
        (class attribute) rank of this `Tier` class (not to be set directly)
    id_regex : str
        (class attribute) regex used to restrict form of `Tier` object ids. Should contain 1 group that captures the id.
    gui_cls : Any
        (class attribute) The class called upon initialisation to use as gui for this object. Constructor should take `self` as first
        argument.
    """

    rank: ClassVar[int] = -1
    id_regex: ClassVar[str] = r"(\d+)"

    gui_cls = BaseTierGui

    hierarchy: ClassVar[List[Type[TierBase]]]

    @cached_class_prop  # type: ignore[no-redef]
    def hierarchy(cls) -> List[Type[TierBase]]:
        """
        Gets the hierarchy from `env.project`.
        """
        assert env.project
        return env.project.hierarchy

    pretty_type: ClassVar[str]

    @cached_class_prop  # type: ignore[no-redef]
    def pretty_type(cls) -> str:
        """
        Name used to display this Tier. Defaults to `cls.__name__`.
        """
        return cls.__name__

    short_type: ClassVar[str]

    @cached_class_prop  # type: ignore[no-redef]
    def short_type(cls) -> str:
        """
        Name used to programmatically refer to instances of this `Tier`.

        Default, take pretty type, make lowercase and remove vowels.
        """
        return cls.pretty_type.lower().translate(str.maketrans(dict.fromkeys("aeiou")))

    name_part_template: ClassVar[str]

    @cached_class_prop  # type: ignore[no-redef]
    def name_part_template(cls) -> str:
        """
        Python template that's filled in with `self.id` to create segment of the `Tier` object's name.
        """
        return cls.pretty_type + "{}"

    name_part_regex: ClassVar[str]

    @cached_class_prop  # type: ignore[no-redef]
    def name_part_regex(cls) -> str:
        """
        Regex where first group matches `id` part of string. Default is fill in `cls.name_part_template` with
        `cls.id_regex`.
        """
        return cls.name_part_template.format(cls.id_regex)

    parent_cls: ClassVar[Union[Type[TierBase], None]]

    @cached_class_prop  # type: ignore[no-redef]
    def parent_cls(cls) -> Union[Type[TierBase], None]:
        """
        `Tier` above this `Tier`, `None` if doesn't have one
        """
        assert env.project

        if cls.rank is None or cls.rank <= 0:
            return None
        return cls.hierarchy[cls.rank - 1]

    child_cls: ClassVar[Union[Type[TierBase], None]]

    @cached_class_prop  # type: ignore[no-redef]
    def child_cls(cls) -> Union[Type[TierBase], None]:
        """
        `Tier` below this `Tier`, `None` if doesn't have one
        """
        assert env.project

        if cls.rank is None or cls.rank >= (len(cls.hierarchy) - 1):
            return None
        return cls.hierarchy[cls.rank + 1]

    default_template: ClassVar[Union[Path, None]]

    @cached_class_prop  # type: ignore[no-redef]
    def default_template(cls) -> Union[Path, None]:
        """
        Template used to render a tier file by default.
        """
        return Path(cls.pretty_type) / f"{cls.pretty_type}.tmplt.ipynb"

    _meta_folder_name: ClassVar[str]

    @cached_class_prop  # type: ignore[no-redef]
    def _meta_folder_name(cls) -> str:
        """
        Form of meta folder name. (Just fills in `config.META_DIR_TEMPLATE` with `cls.short_type`).
        """
        return config.META_DIR_TEMPLATE.format(cls.short_type)

    def __new__(cls, *args: str, **kwargs):
        obj = cls.cache.get(args)
        if obj:
            return obj
        obj = object.__new__(cls)
        cls.cache[args] = obj
        return obj

    @classmethod
    def parse_name(cls, name: str) -> Tuple[str, ...]:
        """
        Ask `env.project` to parse name.
        """
        if not env.project:
            raise RuntimeError(
                "Attempting to parse a name before a project is initialised"
            )
        return env.project.parse_name(name)

    @classmethod
    def _iter_meta_dir(cls, path: Path) -> Iterator[Tuple[str, ...]]:
        for meta_file in os.scandir(path):
            if not meta_file.is_file() or not meta_file.name.endswith(".json"):
                continue
            yield cls.parse_name(meta_file.name[:-5])

    def __init__(self, *args: str):
        self._identifiers = tuple(filter(None, args))
        self.gui = self.gui_cls(self)

        if len(self._identifiers) != self.rank:
            raise ValueError(
                f"Invalid number of identifiers in {self._identifiers}, expecting {self.rank}."
            )

        if self.parse_name(self.name) != self.identifiers:
            raise ValueError(
                f"Invalid identifiers - {self._identifiers}, resulting name ('{self.name}') not in a parsable form "
            )

        if self.meta_file:
            self.meta: Meta = Meta(self.meta_file)

    def setup_files(self, template: Union[Path, None] = None):
        """
        Create all the files needed for a valid `Tier` object to exist.

        This includes its `.ipynb` file, its parent folder, its own folder and its `Meta` file.

        Will render `.ipynb` file as Jinja template engine, passing to the template `self` with names given by
        `self.short_type` and `tier`.

        Parameters
        ----------
        template : Path
            path to template file to render to create `.ipynb` file.
        """
        if template is None:
            template = self.default_template

        if self.exists():
            raise FileExistsError(f"Meta for {self.name} exists already")

        if self.file and self.file.exists():
            raise FileExistsError(f"Notebook for {self.name} exists already")

        print(f"Creating files for {self.name}")

        print(f"Meta ({self.meta_file})")

        with FileMaker() as maker:
            maker.mkdir(self.meta.file.parent, exist_ok=True)
            maker.write_file(self.meta.file, "{}")

            print("Writing Meta Data")

            self.started = datetime.datetime.now()

            print("Success")

            print(f"Creating Tier File ({self.file}) using template ({template})")
            maker.write_file(self.file, self.render_template(template))
            print("Success")

            print(f"Creating Tier Folder ({self.folder})")

            self.folder.mkdir(exist_ok=True)

            print("Success")

        print("All Done")

    description = MetaAttr()
    conclusion = MetaAttr()
    started = MetaAttr(str_to_date, date_to_str)

    @cached_prop
    def identifiers(self) -> Tuple[str, ...]:
        """
        Read only copy of identifiers that make up this `Tier` object.
        """
        return self._identifiers

    @cached_prop
    def name(self) -> str:
        """
        Full name of `Tier` object. Made by concatenating each parent's `self.name_part_template` filled with each parent's
        `self.id`.

        Examples
        --------

            >>> wp = WorkPackage('2')
            >>> exp = Experiment('2', '3')
            >>> smpl = Sample('2', '3', 'c')
            >>> wp.name_part_template.format(wp.id)
            WP2
            >>> exp.name_part_template.format(exp.id)
            .3
            >>> smpl.name_part_template.format(smpl.id)
            c
            >>> smpl.name  # all 3 joined together
            WP2.3c
        """

        return "".join(
            cls.name_part_template.format(id)
            for cls, id in zip(self.hierarchy[1:], self.identifiers)
        )

    @cached_prop
    def id(self) -> str:
        """
        Shortcut for getting final identifier.

        Examples
        --------

            >>> from my_project import project
            >>> smpl = project.env('WP2.3c')
            >>> smpl.identifiers
            ['2', '3', 'c']
            >>> smpl.id
            'c'
        """
        return self._identifiers[-1]

    @cached_prop
    def folder(self) -> Path:
        """
        Path to folder where the contents of this `Tier` lives.

        Defaults to `self.parent.folder / self.name`.
        """
        if self.parent:
            return self.parent.folder / self.name
        else:  # this is bad
            return Path(self.name)

    @cached_prop
    def meta_file(self) -> Path:
        """
        Path to where meta file for this `Tier` object should be.

        Returns
        -------
        meta_file : Path
            Defaults to `self.parent.folder / self._meta_folder_name / (self.name + '.json')`
        """
        assert self.parent
        return self.parent.folder / self._meta_folder_name / (self.name + ".json")

    @cached_prop
    def highlights_file(self) -> Union[Path, None]:
        """
        Path to where highlights file for this `Tier` object should be.

        Returns
        -------
        highlights_file : Path
            Defaults to `self.parent.folder / self._meta_folder_name / (self.name + '.hlts')`
        """
        assert self.parent
        return self.parent.folder / self._meta_folder_name / (self.name + ".hlts")

    @cached_prop
    def cache_file(self) -> Union[Path, None]:
        """
        Path to where cache file for this `Tier` object will be.

        Returns
        -------
        cache_file : Path
            Defaults to `self.parent.folder / self._meta_folder_name / (self.name + '.cache')`
        """
        assert self.parent
        return self.parent.folder / self._meta_folder_name / (self.name + ".cache")

    @cached_prop
    def file(self) -> Union[Path, None]:
        """
        Path to where `.ipynb` file for this `Tier` instance will be.

        Returns
        -------
        file : Path
            Defaults to self.parent.folder / (self.name + '.ipynb').
        """
        assert self.parent
        return self.parent.folder / (self.name + ".ipynb")

    @cached_prop
    def parent(self) -> Union[TierBase, None]:
        """
        Parent of this `Tier` _instance, `None` if has no parent :'(
        """
        if self.parent_cls:
            return self.parent_cls(*self._identifiers[:-1])
        return None

    @cached_prop
    def href(self) -> Union[str, None]:
        """
        href usable in notebook HTML giving link to `self.file`.

        Assumes that `os.getcwd()` reflects the directory of the currently opened `.ipynb` (usually true, unless you're
        changing working dir).

        Does do escaping on the HTML, but is maybe pointless!

        Returns
        -------
        href : str
            href usable in notebook HTML.
        """
        if self.file:
            return html.escape(Path(os.path.relpath(self.file, os.getcwd())).as_posix())
        else:
            return None

    def exists(self) -> bool:
        """
        returns True if this `Tier` object has already been setup (e.g. by `self.setup_files`)
        """
        return self.meta_file.exists()

    def get_child(self, id: str) -> TierBase:
        """
        Get a child according to the given `id`.

        Parameters
        ----------
        id : str
            id to add `self.identifiers` to form new `Tier` object of tier below.

        Returns
        -------
        child : Type[TierBase]
            child `Tier` object.
        """
        assert self.child_cls
        return self.child_cls(*self._identifiers, id)

    def serialize(self):
        data = dict(self.meta)
        data["identifiers"] = self.identifiers
        data["name"] = self.name
        data["file"] = str(self.file)

        parents = []
        parent = self.parent

        while parent:
            parents.append(parent)
            parent = parent.parent

        data["parents"] = [parent.name for parent in parents][::-1]
        data["children"] = [child.name for child in self]

        return data

    def children_df(
        self, *, include: Optional[List[str]] = None, exclude: Optional[List[str]]
    ) -> Union[pd.DataFrame, None]:
        """
        Build an `UnescapedDataFrame` containing rows from each child of this `Tier`. Columns are inferred from contents
        of meta files.

        Parameters
        ----------
        include : Sequence[str]
            names of columns to include in children DataFrame
        exclude : Sequence[str]
            names of columns to drop from children DataFrame

        Notes
        -----
        Parameters include and exclude are mutually exclusive.

        Returns
        -------
        df : UnescapedDataFrame, None
            DataFrame containing `Tier`'s children. If no children then returns `None`.
        """
        if self.child_cls is None:
            return None

        if include and exclude:
            raise ValueError("Only one of include or exclude can be provided")

        children = list(self)

        if not children:
            return None

        data: Dict[str, List[Any]] = defaultdict(list)
        attributes_set: set[str] = set()

        for child in children:
            attributes_set.update(child.meta.keys())
            data["Name"].append(child.name)

        attributes = list(attributes_set)

        for child in children:
            for attr in attributes:
                try:
                    val = getattr(child, attr)
                except AttributeError:
                    val = child.meta.get(attr)
                data[attr].append(val)

        for tier in children:
            data["Obj"].append(tier)

        df = pd.DataFrame(data=data)
        df = df.set_index("Name")
        df = df.sort_values("started", axis=0)

        priority_columns = ["started", "description"]
        if "conclusion" in df.columns:
            priority_columns.append("conclusion")

        df = df[
            priority_columns
            + [col for col in df.columns if col not in priority_columns]
        ]

        if include:
            df = df.loc[:, include]

        if exclude:
            df = df.drop(exclude, axis="columns")

        return cast(pd.DataFrame, df)

    @classmethod
    def get_templates(cls) -> List[Path]:
        """
        Get all the templates for this `Tier`.
        """
        assert env.project

        return [
            Path(cls.pretty_type) / entry.name
            for entry in os.scandir(env.project.template_folder / cls.pretty_type)
            if entry.is_file()
        ]

    def render_template(self, template) -> str:
        """
        Render template file passing `self` as `self.short_type`.

        Parameters
        ----------
        template : Path
            path to template file.

        Returns
        -------
        rendered_text : str
            template rendered with `self`.
        """
        assert env.project

        template = env.project.template_env.get_template(template)
        return template.render(**{self.short_type: self, "tier": self})

    def open_folder(self):
        """
        Open `self.folder` in explorer

        Notes
        -----
        Only works on Windows machines.

        Window is opened via the Jupyter server, not the browser, so if you are not accessing jupyter on localhost then
        the window won't open for you!
        """
        open_file(self.folder)

    def get_highlights(self) -> Union[Dict[str, List[Dict[str, Dict[str, Any]]]], None]:
        """
        Get dictionary of highlights for this `Tier` _instance.

        This dictionary is in a form that can be rendered in the notebook using `IPython.display.publish_display_data`
        see Examples.

        This is implemented for you with `self.display_highlights()`.

        Returns
        -------
        highlights : dict
            Get dictionary of highlights for this `Tier` _instance. If the highlights file doesn't exist, just returns an
            empty dict.

        Examples
        --------

            >>> wp = project['WP1']
            >>> for title, outputs in wp.get_highlights():
            ...     print("Displaying Highlight:", title)
            ...     for output in outputs:
            ...         IPython.display.publish_display_data(**output)
        """
        if self.highlights_file and self.highlights_file.exists():
            highlights = json.loads(self.highlights_file.read_text())
            return highlights
        else:
            return {}

    def add_highlight(
        self, name: str, data: List[Dict[str, Dict[str, Any]]], overwrite: bool = True
    ):
        """
        Add a highlight to `self.highlights_file`.

        This is usually done behind the scenes using the `%%hlt My Title` magic.

        Parameters
        ----------
        name : str
            Name of highlight (also taken as the title).
        data : list
            list of data and metadata that can be passed to `IPython.display.publish_display_data` to render.
        overwrite : bool
            If `False` will raise an exception if a highlight of the same `name` exists. Default is `True`
        """
        highlights = self.get_highlights()

        if self.highlights_file and highlights is not None:
            highlights = highlights.copy()
        else:
            return

        if not overwrite and name in highlights:
            raise KeyError("Attempting to overwrite existing meta value")

        highlights[name] = data
        self.highlights_file.write_text(json.dumps(highlights))

    def remove_highlight(self, name: str):
        """
        Remove highlight from highlight file. Performed by calling `get_highlights()`, then deleting the key `name` from
        the dictionary, then re-writing the highlights... if you're interested!
        """
        highlights = self.get_highlights()

        if not highlights or not self.highlights_file:
            return

        del highlights[name]
        self.highlights_file.write_text(json.dumps(highlights))

    def get_cached(self) -> Union[Dict[str, List[Dict[str, Dict[str, Any]]]], None]:
        """
        Retrieve cached output from `self.cache_file`.

        Keys are a hash of text from cell that created this output.

        Returns
        -------
        cached : dict
            dictionary with all the cached outputs. In the same form as `self.get_highlights()`. Returns an empty dict
            if no `cache_file` exists.
        """
        if self.cache_file and self.cache_file.exists():
            cache = json.loads(self.cache_file.read_text())
            return cache
        else:
            return {}

    def cache_result(
        self, name: str, data: List[Dict[str, Dict[str, Any]]], overwrite: bool = True
    ):
        """
        Cache a result in `self.cache_file`.

        This is usually done behind the scenes using the `%%cache` magic.

        Parameters
        ----------
        name : str
            name used as key in JSON for this cached result. When using `%%cache` this is a hash of the cell text
        data : dict
            list of data and metadata that can be passed to `IPython.display.publish_display_data` to render.
        overwrite : bool
            If `False` will raise an exception if a cached result of the same `name` exists. Default is `True`
        """
        if not self.cache_file:
            return

        cache = self.get_cached()

        if cache:
            cache = cache.copy()
        else:
            return

        if not overwrite and name in cache:
            raise KeyError("Attempting to overwrite existing meta value")

        cache[name] = data
        self.cache_file.write_text(json.dumps(cache))

    def remove_cached(self, name):
        """
        Remove cached output according to the `name` provided.
        """
        cached = self.get_cached()
        del cached[name]
        self.cache_file.write_text(json.dumps(cached))

    def __truediv__(self, other) -> Path:
        return self.folder / other

    def __getitem__(self, item) -> TierBase:
        """
        Equivalent to `self.get_child(item)`.
        """
        return self.get_child(item)

    def __iter__(self) -> Iterator[Any]:
        """
        Iterates over all children (in no particular order). Children are found by looking through the child meta
        folder.

        Empty iterator if no children.
        """
        assert self.child_cls

        child_cls = self.child_cls
        child_meta_dir = self / child_cls._meta_folder_name
        if child_meta_dir.exists():
            return (
                child_cls(*identifiers)
                for identifiers in self._iter_meta_dir(child_meta_dir)
            )
        return iter(())

    def __repr__(self):
        return f'<{self.__class__.__name__} "{self.name}">'

    def _repr_html_(self) -> str:
        block = f'h{len(self.identifiers) + 1} style="display: inline;"'
        return (
            f'<a href="{self.href}"'
            f' target="_blank"><{block}>{html.escape(self.name)}</{block}</a>'
        )

    def __getattr__(self, item):
        if env.project and item in env.project.rank_map:
            rank = env.project.rank_map[item]
            return env.project.hierarchy[rank](*self.identifiers[:rank])
        raise AttributeError(item)

    def remove_files(self):
        """
        Deletes files associated with a `Tier`
        """
        if self.file:
            self.file.unlink()

        if self.meta_file:
            self.meta_file.unlink()
